unit uUpdVer;
{$INCLUDE 'UpdVer.inc'}

interface

uses
  SysUtils, Classes, Windows,
  unitResourceDetails,
  unitResourceVersionInfo,
  unitResFile,
  unitPEFile,
  uConsoleApp,
  uPCRE;

type
  TUpdVer = class(TConsoleApp)
    private
      fVersionString: string;
      fVersionParts: array[0..3] of integer;
      fIncreaseBuild: boolean;
      fStringsToSet: TStringList;
      fCreateBackup: boolean;
      fOutputFileName: string;
      fOutputFileNameUsed: boolean;
      fBinaryVerRegExp: TRegExp;
      fTextVerRegExp: TRegExp;
      fTextValueRegExp: TRegExp;
      function GetBinaryVerRegExp: TRegExp;
      function GetTextVerRegExp: TRegExp;
      function GetStringsToSet: TStrings;
      function GetTextValueRegExp: TRegExp;
    protected
      function AppTitle: string; override;
      function AppVersion: string; override;
      function DoExecute: integer; override;
      procedure ShowSyntax(const Error: string = ''); override;
      procedure ParamInitialize; override;
      function PrepareOutputFileName(const InputFileName: string): string;
      function PrepareOutputFile(const OutputFileName: string): boolean;
      function ProcessResources(Module: TResourceModule): boolean;
      function ProcessResourcesInFile(Module: TResourceModule; const FileName: string): boolean;
      function ProcessFile(const FileName: string): boolean;
      function ProcessPEFile(const FileName: string): boolean;
      function ProcessRESFile(const FileName: string): boolean;
      function ProcessRCFile(const FileName: string): boolean;
      function UpdateVersion(var Version: TULargeInteger): string;
      property BinaryVerRegExp: TRegExp read GetBinaryVerRegExp;
      property TextVerRegExp: TRegExp read GetTextVerRegExp;
      property TextValueRegExp: TRegExp read GetTextValueRegExp;
      property StringsToSet: TStrings read GetStringsToSet;
    public
      destructor Destroy; override;
    end;

const
  RESCODE_NOTHINGTODO = 1;
  RESCODE_BADPARAMS = 2;
  RESCODE_FAILED = 3;
  RESCODE_FILENOTFOUND = 4;
  
implementation

{ TUpdVer }

function TUpdVer.AppTitle: string;
begin
  Result := 'UpdVer';
end;

function TUpdVer.AppVersion: string;
begin
  Result := {$INCLUDE 'UpdVer.version'} ;
end;

procedure TUpdVer.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '<arg> [<arg>] ...'); Writeln;
  Writeln;
  WriteColored(ccWhite, ' -h, -?'); Writeln(' ....... Show this help screen.');
  WriteColored(ccWhite, ' -b[-]'); Writeln(' ........ [Don''t] Increase build number.');
  WriteColored(ccWhite, ' -v a.b'); Writeln(' ....... Set version to a.b.');
  WriteColored(ccWhite, ' -v a.b.c'); Writeln(' ..... Set version to a.b.c.');
  WriteColored(ccWhite, ' -v a.b.c.d'); Writeln(' ... Set version to a.b.c.d.');
  WriteColored(ccWhite, ' -v-'); Writeln(' .......... Don''t set version.');
  WriteColored(ccWhite, ' -f <format>'); Writeln('... Set version string format to <format>. E.g. "%d.%d.%d.%d".');
  WriteColored(ccWhite, ' -s <string>'); Writeln('... Set a version string to a custom value. The format of');
  WriteColored(ccWhite, ''); Writeln('                <string> is <name>=<value> for setting or <name> for');
  WriteColored(ccWhite, ''); Writeln('                deletion, where common <name>s are (case-sensitive):');
  WriteColored(ccWhite, ''); Writeln('                Comments, CompanyName, FileDescription, FileVersion,');
  WriteColored(ccWhite, ''); Writeln('                InternalName, LegalCopyright, LegalTrademarks,');
  WriteColored(ccWhite, ''); Writeln('                OriginalFilename, ProductName, ProductVersion.');
  WriteColored(ccWhite, ' -o-'); Writeln(' .......... Write output over the original file (default).');
  WriteColored(ccWhite, ' -o <file>'); Writeln(' .... Write output to <file>.');
  WriteColored(ccWhite, ' -bak[-]'); Writeln(' ...... [Don''t] Create a backup file.');
  Writeln;
  WriteColored(ccWhite, ' <resfile>'); Writeln(' .... *.res file to update.');
  WriteColored(ccWhite, ' <rcfile>'); Writeln(' ..... *.rc file to update.');
  WriteColored(ccWhite, ' <exefile>'); Writeln(' .... *.exe or other PE file to update.');
  Writeln;
  Writeln;
end;

procedure TUpdVer.ParamInitialize;
var
  i: integer;
begin
  inherited;
  for i := 0 to High(fVersionParts) do
    fVersionParts[i] := -1;
  fIncreaseBuild := False;
  fVersionString := '%d.%d.%d.%d';
  fCreateBackup := True;
  fOutputFileName := '';
  fOutputFileNameUsed := False;
  FreeAndNil(fStringsToSet);
end;

function TUpdVer.DoExecute: integer;
var
  Param: string;
  dot, i: integer;
begin
  if ParamCount = 0 then
  begin
    ShowSyntax;
    Result := RESCODE_OK;
  end
  else
  begin
    Result := RESCODE_NOTHINGTODO;
    ParamInitialize;
    while ParamGetNext(Param) do
      if Param[1] = '-' then
      begin
        if (Param = '-?') or (Param = '-h') then
        begin
          ShowSyntax;
          if Result in [RESCODE_OK, RESCODE_NOTHINGTODO] then
            Result := RESCODE_OK;
          Break;
        end
        else if (Param = '-b') then
          fIncreaseBuild := True
        else if (Param = '-b-') then
          fIncreaseBuild := False
        else if (Param = '-f') then
          if ParamGetNext(Param) then
            fVersionString := Param
          else
          begin
            ShowSyntax('With -f a version format string must be provided.');
            Result := RESCODE_BADPARAMS;
            Break;
          end
        else if (Param = '-s') then
          if ParamGetNext(Param) then
            StringsToSet.Add(Param)
          else
          begin
            ShowSyntax('With -s a modification string must be provided.');
            Result := RESCODE_BADPARAMS;
            Break;
          end
        else if (Param = '-v') then
          if ParamGetNext(Param) then
          begin
            for i := 0 to High(fVersionParts) do
              fVersionParts[i] := -1;
            i := 0;
            while (Param <> '') and (i < Length(fVersionParts)) do
            begin
              dot := Pos('.', Param);
              if dot < 1 then
              begin
                fVersionParts[i] := StrToInt(Param);
                Param := '';
              end
              else
              begin
                fVersionParts[i] := StrToInt(Copy(Param, 1, Pred(dot)));
                System.Delete(Param, 1, dot);
              end;
              Inc(i);
            end;
          end
          else
          begin
            ShowSyntax('With -v a version number must be provided.');
            Result := RESCODE_BADPARAMS;
            Break;
          end
        else if (Param = '-v-') then
        begin
          for i := 0 to High(fVersionParts) do
            fVersionParts[i] := -1;
        end
        else if (Param = '-o') then
          if ParamGetNext(Param) then
          begin
            fOutputFileName := Param;
            fOutputFileNameUsed := False;
          end
          else
          begin
            ShowSyntax('With -o a filename must be provided.');
            Result := RESCODE_BADPARAMS;
            Break;
          end
        else if (Param = '-o-') then
        begin
          fOutputFileName := '';
          fOutputFileNameUsed := False;
        end
        else if (Param = '-bak') then
          fCreateBackup := True
        else if (Param = '-bak-') then
          fCreateBackup := False
        else
        begin
          ShowSyntax('Unknown parameter "%s".', [Param]);
          Result := RESCODE_BADPARAMS;
          Break;
        end
      end
      else
        if FileExists(Param) then
          if (fOutputFileName <> '') and fOutputFileNameUsed then
          begin
            if Result in [RESCODE_OK, RESCODE_NOTHINGTODO] then
              Result := RESCODE_BADPARAMS;
            ShowSyntax('You must provide a separate -o for file "%s".', [Param]);
          end
          else if ProcessFile(Param) then
          begin
            if Result = RESCODE_NOTHINGTODO then
              Result := RESCODE_OK;
          end
          else
            Result := RESCODE_FAILED
        else
        begin
          ShowError('File "%s" does not exist.', [Param]);
          Result := RESCODE_FILENOTFOUND;
        end;
    if Result = RESCODE_NOTHINGTODO then
      ShowError('No files found.');
  end;
end;

function TUpdVer.ProcessFile(const FileName: string): boolean;
var
  Ext: string;
begin
  Result := False;
  if FileExists(FileName) then
  begin
    Write(FileName + ' ... ');
    try
      Ext := UpperCase(ExtractFileExt(FileName));
      if Ext = '.RC' then
        Result := ProcessRCFile(FileName)
      else if Ext = '.RES' then
        Result := ProcessRESFile(FileName)
      else
        Result := ProcessPEFile(FileName);
      if Result then
      begin
        WriteColored(ccWhite, 'OK'); Writeln;
      end;
    except
      on E: Exception do
      begin
        ShowError('Exception %s: %s', [E.ClassName, E.Message]);
        Result := False;
      end;
    end;
  end;
end;

function TUpdVer.ProcessRCFile(const FileName: string): boolean;

  function RegExpToVersion(RE: TRegExp): TULargeInteger;
  var
    sVer: string;
  begin
    Result.QuadPart := 0;
    if RE.SubexpressionByName('VER0', sVer) and (sVer <> '') then
      Result.QuadPart := Result.QuadPart or StrToInt(sVer);
    Result.QuadPart := Result.QuadPart shl 16;
    if RE.SubexpressionByName('VER1', sVer) and (sVer <> '') then
      Result.QuadPart := Result.QuadPart or StrToInt(sVer);
    Result.QuadPart := Result.QuadPart shl 16;
    if RE.SubexpressionByName('VER2', sVer) and (sVer <> '') then
      Result.QuadPart := Result.QuadPart or StrToInt(sVer);
    Result.QuadPart := Result.QuadPart shl 16;
    if RE.SubexpressionByName('VER3', sVer) and (sVer <> '') then
      Result.QuadPart := Result.QuadPart or StrToInt(sVer);
  end;

var
  L: TStringList;
  i: integer;
  Ver: TULargeInteger;
  OutputFileName: string;
  Changed: boolean;
  Name, Value: string;
  j, ix: integer;
begin
  Result := True;
  Changed := False;
  OutputFileName := PrepareOutputFileName(FileName);
  L := TStringList.Create;
  try
    L.LoadFromFile(FileName);
    for i := 0 to Pred(L.Count) do
    begin
      if BinaryVerRegExp.Match(L[i]) then
      begin
        Ver := RegExpToVersion(BinaryVerRegExp);
        UpdateVersion(Ver);
        L[i] := BinaryVerRegExp.SubexpressionByName('PREFIX')
              + BinaryVerRegExp.SubexpressionByName('NAME')
              + BinaryVerRegExp.SubexpressionByName('SEPARATOR')
              +       IntToStr((Ver.HighPart shr 16) and $ffff)
              + ',' + IntToStr((Ver.HighPart       ) and $ffff)
              + ',' + IntToStr((Ver.LowPart  shr 16) and $ffff)
              + ',' + IntToStr((Ver.LowPart        ) and $ffff)
              + BinaryVerRegExp.SubexpressionByName('SUFFIX')
              ;
        Changed := True;
      end
      else if TextVerRegExp.Match(L[i]) then
      begin
        Ver := RegExpToVersion(TextVerRegExp);
        L[i] := TextVerRegExp.SubexpressionByName('PREFIX')
              + TextVerRegExp.SubexpressionByName('NAME')
              + TextVerRegExp.SubexpressionByName('SEPARATOR')
              + UpdateVersion(Ver)
              + TextVerRegExp.SubexpressionByName('SUFFIX')
              ;
        Changed := True;
      end;
      if Assigned(fStringsToSet) then
      begin
        if TextValueRegExp.Match(L[i]) then
        begin
          for j := 0 to Pred(StringsToSet.Count) do
          begin
            Name := StringsToSet[j];
            ix := Pos('=', Name);
            if ix > 0 then
            begin
              Value := Copy(Name, Succ(ix), MaxInt);
              SetLength(Name, Pred(ix));
              if Name = TextValueRegExp.SubexpressionByName('NAME') then
              begin
                L[i] := TextValueRegExp.SubexpressionByName('PREFIX')
                      + Name
                      + TextValueRegExp.SubexpressionByName('SEPARATOR')
                      + Value
                      + TextValueRegExp.SubexpressionByName('SUFFIX')
                      ;
                Changed := True;
              end;
            end
            else
            begin
              if Name = TextValueRegExp.SubexpressionByName('NAME') then
              begin
                L[i] := '//' + L[i];
                Changed := True;
              end;
            end;
          end;
        end;
      end;
    end;
    if Changed then
      if PrepareOutputFile(OutputFileName) then
      begin
        L.SaveToFile(OutputFileName);
        Result := True;
      end
      else
        Result := False;
  finally
    FreeAndNil(L);
    end;
end;

function TUpdVer.ProcessResources(Module: TResourceModule): boolean;
var
  VerInfo: TVersionInfoResourceDetails;
  Version: TULargeInteger;
  VersionStr: string;
  i, j: integer;
  Name, Value: string;
  ix: integer;
begin
  Result := False;
  for i := 0 to Pred(Module.ResourceCount) do
  begin
    if Module.ResourceDetails[i] is TVersionInfoResourceDetails then
    begin
      VerInfo := TVersionInfoResourceDetails(Module.ResourceDetails[i]);
      Version := VerInfo.FileVersion;
      VersionStr := UpdateVersion(Version);
      VerInfo.FileVersion := Version;
      VerInfo.ProductVersion := Version;
      // Update version strings
      for j := 0 to Pred(VerInfo.KeyCount) do
      begin
        if (VerInfo.Key[j].KeyName = 'FileVersion') or (VerInfo.Key[j].KeyName = 'ProductVersion') then
          VerInfo.SetKeyValue(VerInfo.Key[j].KeyName, VersionStr);
      end;
      // Update regular strings
      if Assigned(fStringsToSet) then
      begin
        for j := 0 to Pred(StringsToSet.Count) do
        begin
          Name := StringsToSet[j];
          ix := Pos('=', Name);
          if ix > 0 then
          begin
            Value := Copy(Name, Succ(ix), MaxInt);
            SetLength(Name, Pred(ix));
            VerInfo.SetKeyValue(Name, Value);
          end
          else
          begin
            ix := VerInfo.IndexOf(Name);
            if ix >= 0 then
              VerInfo.DeleteKey(ix);
          end;
        end;
      end;
      Result := True;
    end;
  end;
end;

function TUpdVer.ProcessResourcesInFile(Module: TResourceModule; const FileName: string): boolean;
var
  OutputFileName: string;
begin
  Result := False;
  OutputFileName := PrepareOutputFileName(FileName);
  Module.LoadFromFile(FileName);
  if ProcessResources(Module) then
  begin
    if PrepareOutputFile(OutputFileName) then
    begin
      Module.SaveToFile(OutputFileName);
      Result := True;
    end
    else
      Result := False;
  end;
end;

function TUpdVer.ProcessPEFile(const FileName: string): boolean;
var
  Module: TPEResourceModule;
begin
  Module := TPEResourceModule.Create;
  try
    Result := ProcessResourcesInFile(Module, FileName);
  finally
    FreeAndNil(Module);
  end;
end;

function TUpdVer.ProcessRESFile(const FileName: string): boolean;
var
  Module: TResModule;
begin
  Module := TResModule.Create;
  try
    Result := ProcessResourcesInFile(Module, FileName);
  finally
    FreeAndNil(Module);
  end;
end;

destructor TUpdVer.Destroy;
begin
  FreeAndNil(fStringsToSet);
  RegExFreeAndNil(fBinaryVerRegExp);
  RegExFreeAndNil(fTextVerRegExp);
  RegExFreeAndNil(fTextValueRegExp);
  inherited;
end;

function TUpdVer.GetBinaryVerRegExp: TRegExp;
begin
  if fBinaryVerRegExp = nil then
    fBinaryVerRegExp := RegExCreate('^(?P<PREFIX>\s*)(?P<NAME>FILEVERSION|PRODUCTVERSION)(?P<SEPARATOR>\s+)(?P<VERSION>(?P<VER0>\d+)\s*,\s*(?P<VER1>\d+)\s*,\s*(?P<VER2>\d+)\s*,\s*(?P<VER3>\d+))(?P<SUFFIX>\s*)$');
  Result := fBinaryVerRegExp;
end;

function TUpdVer.GetTextVerRegExp: TRegExp;
begin
  if fTextVerRegExp = nil then
    fTextVerRegExp := RegExCreate('^(?P<PREFIX>\s*VALUE\s+")(?P<NAME>FileVersion|ProductVersion)(?P<SEPARATOR>"\s*,\s*")(?P<VERSION>(?P<VER0>\d+)(?:\.(?P<VER1>\d+)(?:\.(?P<VER2>\d+)(?:\.(?P<VER3>\d+))?)?)?)(?P<SUFFIX>"\s*)$');
  Result := fTextVerRegExp;
end;

function TUpdVer.UpdateVersion(var Version: TULargeInteger): string;

  procedure UpdateHigh(var Item: DWORD; Value: integer);
  begin
    Item := (Item and $0000ffff) or ((DWORD(Value) and $ffff) shl 16);
  end;

  procedure UpdateLow(var Item: DWORD; Value: integer);
  begin
    Item := (Item and $ffff0000) or (DWORD(Value) and $ffff);
  end;

begin
  if fVersionParts[0] >= 0 then
    UpdateHigh(Version.HighPart, fVersionParts[0]);
  if fVersionParts[1] >= 0 then
    UpdateLow(Version.HighPart, fVersionParts[1]);
  if fVersionParts[2] >= 0 then
    UpdateHigh(Version.LowPart, fVersionParts[2]);
  if fVersionParts[3] >= 0 then
    UpdateLow(Version.LowPart, fVersionParts[3]);
  if fIncreaseBuild then
    UpdateLow(Version.LowPart, Succ(Version.LowPart and $ffff));
  Result := Format(fVersionString, [Version.HighPart shr 16, Version.HighPart and $0000ffff, Version.LowPart shr 16, Version.LowPart and $0000ffff]);
end;

function TUpdVer.GetStringsToSet: TStrings;
begin
  if fStringsToSet = nil then
    fStringsToSet := TStringList.Create;
  Result := fStringsToSet;
end;

function TUpdVer.GetTextValueRegExp: TRegExp;
begin
  if fTextValueRegExp = nil then
    fTextValueRegExp := RegExCreate('^(?P<PREFIX>\s*VALUE\s+")(?P<NAME>[^"]+)(?P<SEPARATOR>"\s*,\s*")(?P<VALUE>[^"]*)(?P<SUFFIX>"\s*)$');
  Result := fTextValueRegExp;
end;

function TUpdVer.PrepareOutputFileName(const InputFileName: string): string;
begin
  if fOutputFileName <> '' then
    if fOutputFileNameUsed then
      Raise Exception.CreateFmt('You must provide a separate -o for file "%s".', [InputFileName])
    else
    begin
      Result := fOutputFileName;
      fOutputFileNameUsed := True;
    end
  else
    Result := InputFileName;
end;

function TUpdVer.PrepareOutputFile(const OutputFileName: string): boolean;
begin
  Result := False;
  if fCreateBackup then
  begin
    if FileExists(OutputFileName + '.bak') then
      SysUtils.DeleteFile(OutputFileName + '.bak');
    if FileExists(OutputFileName) then
      RenameFile(OutputFileName, OutputFileName + '.bak');
    if not FileExists(OutputFileName) then
      Result := True;
  end
  else
    Result := True;
end;

end.
